from bandicoot.helper.tools import OrderedDict, warning_str
from bandicoot.helper.group import group_records
import bandicoot as bc

from functools import partial

try:
    next
except NameError:
    from bandicoot.helper.fixes import next


def flatten(d, parent_key='', separator='__'):
    """
    Flatten a nested dictionary.

    Parameters
    ----------
    d: dict_like
        Dictionary to flatten.
    parent_key: string, optional
        Concatenated names of the parent keys.
    separator: string, optional
        Separator between the names of the each key.
        The default separator is '_'.

    Examples
    --------

    >>> d = {'alpha': 1, 'beta': {'a': 10, 'b': 42}}
    >>> flatten(d) == {'alpha': 1, 'beta_a': 10, 'beta_b': 42}
    True
    >>> flatten(d, separator='.') == {'alpha': 1, 'beta.a': 10, 'beta.b': 42}
    True

    """
    items = []
    for k, v in d.items():
        new_key = parent_key + separator + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten(v, new_key, separator).items())
        else:
            items.append((new_key, v))
    return OrderedDict(items)


def all(user, groupby='week', summary='default', split_week=False, split_day=False, attributes=True, flatten=False):
    """
    Returns a dictionary containing all bandicoot indicators for the user,
    as well as reporting variables.

    The reporting variables include:

    * the path of files containting the antennas and attributes,
    * the current version of bandicoot,
    * the *groupby* method (``'week'`` or ``None``) and the day/night, weekday/weekend filters,
    * the date and time for the first and last records,
    * the range of hours used to detect interactions at night, and the weekend range,
    * the number of bins if the records are grouped weekly,
    * the binary properties ``has_call``, ``has_text``, ``has_home``,
    * the percentage of records missing antennas, and antennas missing (lat, lon) locations,
    * the percentage of contacts not in the network, as well as interactions (for calls, texts, and call durations),
    * the total number of records for the user

    We also include a last set of reporting variables, for the records ignored
    at the loading, due to faulty or incorrect values:

    .. code-block:: python

        {
            'all': 0,
            'interaction': 0,
            'direction': 0,
            'correspondent_id': 0,
            'datetime': 0,
            'call_duration': 0
        }

    with the total number of records ignored (key ``'all'``), as well as the
    number of records with faulty values for each columns.
    """

    # Warn the user if they are selecting weekly and there's only one week
    if groupby == 'week':
        if len(set(r.datetime.isocalendar()[:2] for r in user.records)) <= 1:
            print warning_str('Grouping by week, but all data is from the same week!')

    scalar_type = 'distribution_scalar' if groupby == 'week' else 'scalar'
    summary_type = 'distribution_summarystats' if groupby == 'week' else 'summarystats'

    number_of_interactions_in = partial(bc.individual.number_of_interactions, direction='in')
    number_of_interactions_in.__name__ = 'number_of_interaction_in'
    number_of_interactions_out = partial(bc.individual.number_of_interactions, direction='out')
    number_of_interactions_out.__name__ = 'number_of_interaction_out'

    functions = [
        (bc.individual.active_days, scalar_type),
        (bc.individual.number_of_contacts, scalar_type),
        (bc.individual.call_duration, summary_type),
        (bc.individual.percent_nocturnal, scalar_type),
        (bc.individual.percent_initiated_conversations, scalar_type),
        (bc.individual.percent_initiated_interactions, scalar_type),
        (bc.individual.response_delay_text, summary_type),
        (bc.individual.response_rate_text, scalar_type),
        (bc.individual.entropy_of_contacts, scalar_type),
        (bc.individual.balance_of_contacts, summary_type),
        (bc.individual.interactions_per_contact, summary_type),
        (bc.individual.interevent_time, summary_type),
        (bc.individual.percent_pareto_interactions, scalar_type),
        (bc.individual.percent_pareto_durations, scalar_type),
        (bc.individual.number_of_interactions, scalar_type),
        (number_of_interactions_in, scalar_type),
        (number_of_interactions_out, scalar_type),
        (bc.spatial.number_of_antennas, scalar_type),
        (bc.spatial.entropy_of_antennas, scalar_type),
        (bc.spatial.percent_at_home, scalar_type),
        (bc.spatial.radius_of_gyration, scalar_type),
        (bc.spatial.frequent_antennas, scalar_type)
    ]

    groups = [[r for r in g] for g in group_records(user, groupby=groupby)]

    reporting = OrderedDict([
        ('antennas_path', user.antennas_path),
        ('attributes_path', user.attributes_path),
        ('version', bc.__version__),
        ('groupby', groupby),
        ('split_week', split_week),
        ('split_day', split_day),
        ('start_time', user.start_time and str(user.start_time)),
        ('end_time', user.end_time and str(user.end_time)),
        ('night_start', str(user.night_start)),
        ('night_end', str(user.night_end)),
        ('weekend', user.weekend),
        ('bins', len(groups)),
        ('has_call', user.has_call),
        ('has_text', user.has_text),
        ('has_home', user.has_home),
        ('percent_records_missing_location', bc.helper.tools.percent_records_missing_location(user)),
        ('antennas_missing_locations', bc.helper.tools.antennas_missing_locations(user)),
        ('percent_outofnetwork_calls', user.percent_outofnetwork_calls),
        ('percent_outofnetwork_texts', user.percent_outofnetwork_texts),
        ('percent_outofnetwork_contacts', user.percent_outofnetwork_contacts),
        ('percent_outofnetwork_call_durations', user.percent_outofnetwork_call_durations),
    ])

    if user.records is not None:
        reporting['number_of_records'] = len(user.records)
    else:
        reporting['number_of_records'] = 0.

    if user.ignored_records is not None:
        reporting['ignored_records'] = user.ignored_records

    returned = OrderedDict([
        ('name', user.name),
        ('reporting', reporting)
    ])

    for fun, datatype in functions:
        try:
            metric = fun(user, groupby=groupby, summary=summary, datatype=datatype, split_week=split_week, split_day=split_day)
        except ValueError:
            metric = fun(user, groupby=groupby, datatype=datatype, split_week=split_week, split_day=split_day)

        returned[fun.__name__] = metric

    if attributes and user.attributes != {}:
        returned['attributes'] = user.attributes

    if flatten is True:
        return globals()['flatten'](returned)

    return returned



"""
Test the bandicoot.utils module, mostly using the numpy library.
"""

import bandicoot as bc
import unittest
from scipy import stats
import numpy as np
import os


class TestUtils(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dir_changed = False

    def setUp(self):
        if not TestUtils._dir_changed:
            abspath = os.path.abspath(__file__)
            name = abspath.index(os.path.basename(__file__))
            abspath = abspath[:name]
            os.chdir(abspath)
            TestUtils._dir_changed = True

        self.user = bc.io.read_orange("samples/u_test.csv", describe=False)

        self.list_1 = np.random.randint(-1000, 1000, size=10000)
        self.list_2 = np.random.randint(1, 1000, size=9001)

    def test_flatten(self):
        d = {'alpha': 1, 'beta': {'a': 10, 'b': 42}}
        self.assertEqual(bc.utils.flatten(d), {
            'alpha': 1,
            'beta__a': 10,
            'beta__b': 42})
        self.assertEqual(bc.utils.flatten(d, separator='.'), {
            'alpha': 1,
            'beta.a': 10,
            'beta.b': 42})

    def test_skewness(self):
        self.assertEqual(bc.helper.tools.skewness([]), None)
        self.assertAlmostEqual(bc.helper.tools.skewness([1, 2, 3, 4, 7]), stats.skew([1, 2, 3, 4, 7]))

    def test_kurtosis(self):
        self.assertEqual(bc.helper.tools.kurtosis([]), None)
        self.assertAlmostEqual(bc.helper.tools.kurtosis([1, 2, 3, 4, 5]), stats.kurtosis([1, 2, 3, 4, 5], fisher=False))
        self.assertAlmostEqual(bc.helper.tools.kurtosis([1, 6, 6, 6, 9, 17]), stats.kurtosis([1, 6, 6, 6, 9, 17], fisher=False))

        self.assertAlmostEqual(bc.helper.tools.kurtosis(self.list_1), stats.kurtosis(self.list_1, fisher=False))
        self.assertAlmostEqual(bc.helper.tools.kurtosis(self.list_2), stats.kurtosis(self.list_2, fisher=False))

    def test_mean(self):
        self.assertEqual(bc.helper.tools.mean([]), None)
        self.assertAlmostEqual(bc.helper.tools.mean(self.list_1), np.average(self.list_1))
        self.assertAlmostEqual(bc.helper.tools.mean(self.list_2), np.average(self.list_2))

    def test_std(self):
        self.assertEqual(bc.helper.tools.std([]), None)
        self.assertAlmostEqual(bc.helper.tools.std(self.list_1), np.std(self.list_1))
        self.assertAlmostEqual(bc.helper.tools.std(self.list_2), np.std(self.list_2))

    def test_median(self):
        self.assertEqual(bc.helper.tools.median([]), None)
        self.assertEqual(bc.helper.tools.median(self.list_1), np.median(self.list_1))
        self.assertEqual(bc.helper.tools.median(self.list_2), np.median(self.list_2))

    def test_entropy(self):
        self.assertEqual(bc.helper.tools.entropy([]), None)
        self.assertAlmostEqual(bc.helper.tools.entropy(self.list_2), stats.entropy(self.list_2))

    def test_great_circle_distance(self):
        pt1 = [-1, -1]
        pt2 = [1, 1]
        pt3 = [0, 0]
        pt4 = [100, -100]
        self.assertEqual(bc.helper.tools.great_circle_distance(pt1, pt2), 0)
        self.assertEqual(bc.helper.tools.great_circle_distance(pt1, pt3), 0)
        self.assertEqual(bc.helper.tools.great_circle_distance(pt1, pt4), 20015.086796020572)
        self.assertEqual(bc.helper.tools.great_circle_distance(pt2, pt3), 0)
        self.assertEqual(bc.helper.tools.great_circle_distance(pt2, pt4), 20015.086796020572)
        self.assertEqual(bc.helper.tools.great_circle_distance(pt3, pt4), 20015.086796020572)

    def test_summary_stats(self):
        self.assertEqual(bc.helper.tools.summary_stats([1, 5, 2]),
            bc.helper.tools.SummaryStats(mean=2.6666666666666665, std=1.699673171197595, min=1.0, max=5.0, median=2.0, skewness=0.5280049792181879, kurtosis=1.4999999999999998, distribution=[1, 2, 5]))
        self.assertEqual(bc.helper.tools.summary_stats([1, 2, 3]),
            bc.helper.tools.SummaryStats(mean=2.0, std=0.816496580927726, min=1.0, max=3.0, median=2.0, skewness=0.0, kurtosis=1.5, distribution=[1, 2, 3]))
        self.assertEqual(bc.helper.tools.summary_stats([]),
            bc.helper.tools.SummaryStats(mean=0., std=0., min=0., max=0., median=0., skewness=0., kurtosis=0., distribution=[]))

    def test_all(self):
        pass



""" Provides compatibility fixes for Jython
"""

__all__ = ['Counter', 'next']


#
# The Counter class is licensed under the MIT Licence (MIT)
# Copyright (c) 2009 Raymond Hettinger
# http://code.activestate.com/recipes/576611-counter-class/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


try:
    from collections import Counter
except ImportError:
    from operator import itemgetter
    from heapq import nlargest
    from itertools import repeat, ifilter

    class Counter(dict):
        '''Dict subclass for counting hashable objects.  Sometimes called a bag
        or multiset.  Elements are stored as dictionary keys and their counts
        are stored as dictionary values.

        >>> Counter('zyzygy')
        Counter({'y': 3, 'z': 2, 'g': 1})

        '''

        def __init__(self, iterable=None, **kwds):
            '''Create a new, empty Counter object.  And if given, count elements
            from an input iterable.  Or, initialize the count from another mapping
            of elements to their counts.

            >>> c = Counter()                           # a new, empty counter
            >>> c = Counter('gallahad')                 # a new counter from an iterable
            >>> c = Counter({'a': 4, 'b': 2})           # a new counter from a mapping
            >>> c = Counter(a=4, b=2)                   # a new counter from keyword args

            '''
            self.update(iterable, **kwds)

        def __missing__(self, key):
            return 0

        def most_common(self, n=None):
            '''List the n most common elements and their counts from the most
            common to the least.  If n is None, then list all element counts.

            >>> Counter('abracadabra').most_common(3)
            [('a', 5), ('r', 2), ('b', 2)]

            '''
            if n is None:
                return sorted(self.iteritems(), key=itemgetter(1), reverse=True)
            return nlargest(n, self.iteritems(), key=itemgetter(1))

        def elements(self):
            '''Iterator over elements repeating each as many times as its count.

            >>> c = Counter('ABCABC')
            >>> sorted(c.elements())
            ['A', 'A', 'B', 'B', 'C', 'C']

            If an element's count has been set to zero or is a negative number,
            elements() will ignore it.

            '''
            for elem, count in self.iteritems():
                for _ in repeat(None, count):
                    yield elem

        # Override dict methods where the meaning changes for Counter objects.

        @classmethod
        def fromkeys(cls, iterable, v=None):
            raise NotImplementedError(
                'Counter.fromkeys() is undefined.  Use Counter(iterable) instead.')

        def update(self, iterable=None, **kwds):
            '''Like dict.update() but add counts instead of replacing them.

            Source can be an iterable, a dictionary, or another Counter instance.

            >>> c = Counter('which')
            >>> c.update('witch')           # add elements from another iterable
            >>> d = Counter('watch')
            >>> c.update(d)                 # add elements from another counter
            >>> c['h']                      # four 'h' in which, witch, and watch
            4

            '''
            if iterable is not None:
                if hasattr(iterable, 'iteritems'):
                    if self:
                        self_get = self.get
                        for elem, count in iterable.iteritems():
                            self[elem] = self_get(elem, 0) + count
                    else:
                        dict.update(self, iterable) # fast path when counter is empty
                else:
                    self_get = self.get
                    for elem in iterable:
                        self[elem] = self_get(elem, 0) + 1
            if kwds:
                self.update(kwds)

        def copy(self):
            'Like dict.copy() but returns a Counter instance instead of a dict.'
            return Counter(self)

        def __delitem__(self, elem):
            'Like dict.__delitem__() but does not raise KeyError for missing values.'
            if elem in self:
                dict.__delitem__(self, elem)

        def __repr__(self):
            if not self:
                return '%s()' % self.__class__.__name__
            items = ', '.join(map('%r: %r'.__mod__, self.most_common()))
            return '%s({%s})' % (self.__class__.__name__, items)

        # Multiset-style mathematical operations discussed in:
        #       Knuth TAOCP Volume II section 4.6.3 exercise 19
        #       and at http://en.wikipedia.org/wiki/Multiset
        #
        # Outputs guaranteed to only include positive counts.
        #
        # To strip negative and zero counts, add-in an empty counter:
        #       c += Counter()

        def __add__(self, other):
            '''Add counts from two counters.

            >>> Counter('abbb') + Counter('bcc')
            Counter({'b': 4, 'c': 2, 'a': 1})


            '''
            if not isinstance(other, Counter):
                return NotImplemented
            result = Counter()
            for elem in set(self) | set(other):
                newcount = self[elem] + other[elem]
                if newcount > 0:
                    result[elem] = newcount
            return result

        def __sub__(self, other):
            ''' Subtract count, but keep only results with positive counts.

            >>> Counter('abbbc') - Counter('bccd')
            Counter({'b': 2, 'a': 1})

            '''
            if not isinstance(other, Counter):
                return NotImplemented
            result = Counter()
            for elem in set(self) | set(other):
                newcount = self[elem] - other[elem]
                if newcount > 0:
                    result[elem] = newcount
            return result

        def __or__(self, other):
            '''Union is the maximum of value in either of the input counters.

            >>> Counter('abbb') | Counter('bcc')
            Counter({'b': 3, 'c': 2, 'a': 1})

            '''
            if not isinstance(other, Counter):
                return NotImplemented
            _max = max
            result = Counter()
            for elem in set(self) | set(other):
                newcount = _max(self[elem], other[elem])
                if newcount > 0:
                    result[elem] = newcount
            return result

        def __and__(self, other):
            ''' Intersection is the minimum of corresponding counts.

            >>> Counter('abbb') & Counter('bcc')
            Counter({'b': 1})

            '''
            if not isinstance(other, Counter):
                return NotImplemented
            _min = min
            result = Counter()
            if len(self) < len(other):
                self, other = other, self
            for elem in ifilter(self.__contains__, other):
                newcount = _min(self[elem], other[elem])
                if newcount > 0:
                    result[elem] = newcount
            return result


class Throw(object): pass
throw = Throw() # easy sentinel hack


def next(iterator, default=throw):
    try:
        iternext = iterator.next.__call__
    except AttributeError:
        raise TypeError("%s object is not an iterator" % type(iterator).__name__)
    try:
        return iternext()
    except StopIteration:
        if default is throw:
            raise
        return default


# Backport of datetime.datetime.totalseconds()
def total_seconds(td):
    return float((td.microseconds +
                  (td.seconds + td.days * 24 * 3600) * 10**6)) / 10**6



import json
import filecmp
import bandicoot as bc
import numpy as np


def _convert(input):
    """
    Recursive convertion of unicode
    """
    if isinstance(input, dict):
        return {_convert(key): _convert(value) for key, value in input.iteritems()}
    elif isinstance(input, list):
        return [_convert(element) for element in input]
    elif isinstance(input, unicode):
        return input.encode('utf-8')
    else:
        return input


def parse_dict(path):
    data = open(path)
    dict_data = _convert(json.load(data))
    return dict_data


def file_equality(f1, f2):
    '''Returns true if the files are the same.'''
    return filecmp.cmp(f1, f2)


def metric_suite(user, answers, decimal=7, **kwargs):
    '''
    Runs the complete metric suite.
    If any of the metrics is different than the expected answer, return False.
    '''

    results = bc.utils.all(user, **kwargs)
    test_result, msg = compare_dict(answers, results, decimal=decimal)
    return test_result, msg


def compare_dict(answer, result, decimal=7):
    '''Returns true if two dictionaries are approximately equal. Returns false otherwise.'''
    flat_answer = bc.utils.flatten(answer)
    flat_result = bc.utils.flatten(result)

    for key in flat_answer.keys():
        if key not in flat_result.keys():
            return False, "The key {} was not there.".format(key)

        answer_v, result_v = flat_answer[key], flat_result[key]

        if isinstance(answer_v, (float, int)) and isinstance(result_v, (float, int)):
            try:
                np.testing.assert_almost_equal(answer_v, result_v, decimal=decimal)
            except AssertionError:
                err_msg = "TEST A The key {} produced a different result: expected {}, got {}.".format(key, answer_v, result_v)
                print False, err_msg
        elif answer_v != result_v:
            print False, "TEST B The key {} produced a different result: expected {}, got {}.".format(key, answer_v, result_v)

    return True, ""



import bandicoot as bc
import unittest
from testing_tools import parse_dict, file_equality, metric_suite, compare_dict
import tempfile
import os
from collections import OrderedDict as OD

class TestExport(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dir_changed = False

    def setUp(self):
        if not TestExport._dir_changed:
            abspath = os.path.abspath(__file__)
            name = abspath.index(os.path.basename(__file__))
            abspath = abspath[:name]
            os.chdir(abspath)
            TestExport._dir_changed = True

    def test_different_keys(self):
        tmp_file = tempfile.NamedTemporaryFile()
        dict1 = {"x": 1, "y": 2, "z": 3}
        dict1 = OD(sorted(dict1.items(), key=lambda t: t[0]))

        dict2 = {"a": 4, "b": 5, "c": 6}
        dict2 = OD(sorted(dict2.items(), key=lambda t: t[0]))

        bc.io.to_csv([dict1, dict2], tmp_file.name)
        self.assertTrue(file_equality(tmp_file.name, "samples/to_csv_different_keys.csv"))

    def test_same_keys(self):
        tmp_file = tempfile.NamedTemporaryFile()
        dict1 = {"a": 1, "b": 2, "c": 3}
        dict1 = OD(sorted(dict1.items(), key=lambda t: t[0]))

        dict2 = {"a": 4, "b": 5, "c": 6}
        dict2 = OD(sorted(dict2.items(), key=lambda t: t[0]))

        bc.io.to_csv([dict1, dict2], tmp_file.name)
        self.assertTrue(file_equality(tmp_file.name, "samples/to_csv_same_keys.csv"))

    def test_different_json(self):
        tmp_file = tempfile.NamedTemporaryFile()
        dict1 = {"name": "dict1", "x": 1, "y": 2, "z": 3}
        dict1 = OD(sorted(dict1.items(), key=lambda t: t[0]))

        dict2 = {"name": "dict2", "a": 4, "b": 5, "c": 6}
        dict2 = OD(sorted(dict2.items(), key=lambda t: t[0]))

        bc.io.to_json([dict1, dict2], tmp_file.name)
        self.assertTrue(file_equality(tmp_file.name, "samples/to_json_different_keys.json"))

    def test_different_json(self):
        tmp_file = tempfile.NamedTemporaryFile()
        dict1 = {"name": "dict1", "a": 1, "b": 2, "c": 3}
        dict1 = OD(sorted(dict1.items(), key=lambda t: t[0]))

        dict2 = {"name": "dict2", "a": 4, "b": 5, "c": 6}
        dict2 = OD(sorted(dict2.items(), key=lambda t: t[0]))

        bc.io.to_json([dict1, dict2], tmp_file.name)
        self.assertTrue(file_equality(tmp_file.name, "samples/to_json_same_keys.json"))



import unittest
import bandicoot as bc
from bandicoot.tests.generate_user import random_burst

from datetime import timedelta
from bandicoot.helper.fixes import total_seconds


class InterEventTests(unittest.TestCase):
    def setUp(self):
        self.user = bc.User()

    def test_empty_interevent(self):
        self.user.records = []
        interevent = bc.individual.interevent_time(self.user)
        self.assertEqual(interevent['allweek']['allday']['text'].values(), [None] * 2)

    def test_sequence_interevent(self):
        # 10 minutes burst
        _delta = timedelta(hours=2, minutes=13)
        self.user.records = random_burst(100, interaction='text',
                                         delta=_delta)
        interevent = bc.individual.interevent_time(self.user, groupby=None)
        self.assertEqual(interevent['allweek']['allday']['text']['mean'],
                         total_seconds(_delta))
        self.assertEqual(interevent['allweek']['allday']['text']['std'], 0)

        _delta = timedelta(hours=10, minutes=10)
        self.user.records = random_burst(100, interaction='text',
                                         delta=_delta)
        interevent = bc.individual.interevent_time(self.user, groupby=None)
        self.assertEqual(interevent['allweek']['allday']['text']['mean'],
                         total_seconds(_delta))



__all__ = ['core', 'individual', 'spatial', 'network', 'helper', 'io', 'utils', 'tests']

from .io import read_csv
from .core import User
from . import individual, spatial, network, helper, utils, io, tests, core

__version__ = "0.3.0"



"""
Contains tools for processing files (reading and writing csv and json files).
"""

from __future__ import with_statement, division

from bandicoot.helper.tools import OrderedDict

from bandicoot.core import User, Record, Position
from bandicoot.helper.tools import percent_records_missing_location, antennas_missing_locations, warning_str
from bandicoot.utils import flatten

from datetime import datetime
from json import dumps
from collections import Counter
import csv
import sys
import os


def to_csv(objects, filename, digits=5):
    """
    Export the flatten indicators of one or several users to CSV.

    Parameters
    ----------
    objects : list
        List of objects to be exported.
    filename : string
        File to export to.
    digits : int
        Precision of floats.

    Examples
    --------
    This function can be use to export the results of :meth`bandicoot.utils.all`.
    >>> U_1 = bc.User()
    >>> U_2 = bc.User()
    >>> bc.to_csv([bc.utils.all(U_1), bc.utils.all(U_2)], 'results_1_2.csv')

    If you only have one object, you can simply pass it as argument:
    >>> bc.to_csv(bc.utils.all(U_1), 'results_1.csv')
    """

    if not isinstance(objects, list):
        objects = [objects]

    data = [flatten(obj) for obj in objects]
    all_keys = [d for datum in data for d in datum.keys()]
    field_names = sorted(set(all_keys), key=lambda x: all_keys.index(x))

    with open(filename, 'wb') as f:
        w = csv.writer(f)
        w.writerow(field_names)

        def make_repr(item):
            if item is None:
                return None
            elif isinstance(item, float):
                return repr(round(item, digits))
            else:
                return str(item)

        for row in data:
            row = dict((k, make_repr(v)) for k, v in row.items())
            w.writerow([make_repr(row.get(k, None)) for k in field_names])

    print "Successfully exported %d object(s) to %s" % (len(objects), filename)


def to_json(objects, filename):
    """
    Export the indicators of one or several users to JSON.

    Parameters
    ----------
    objects : list
        List of objects to be exported.
    filename : string
        File to export to.

    Examples
    --------
    This function can be use to export the results of :meth`bandicoot.utils.all`.
    >>> U_1 = bc.User()
    >>> U_2 = bc.User()
    >>> bc.to_json([bc.utils.all(U_1), bc.utils.all(U_2)], 'results_1_2.json')

    If you only have one object, you can simply pass it as argument:
    >>> bc.to_json(bc.utils.all(U_1), 'results_1.json')
    """

    if not isinstance(objects, list):
        objects = [objects]

    obj_dict = {obj['name']: obj for obj in objects}

    with open(filename, 'wb') as f:
        f.write(dumps(obj_dict, sort_keys=True, indent=4, separators=(',', ': ')))
    print "Successfully exported %d object(s) to %s" % (len(objects), filename)

def _parse_date(string):
    try:
        return datetime.strptime(string, "%Y-%m-%d %H:%M:%S")
    except ValueError as ex:
        return ex


def _parse_record(data):

    _map_duration = lambda s: int(s) if s != '' else None

    def _map_position(data):
        antenna = Position()
        if 'antenna_id' in data:
            antenna.antenna = data['antenna_id']
            return antenna
        if 'latitude' in data and 'longitude' in data:
            antenna.position = float(data['latitude']), float(data['longitude'])
        return antenna

    return Record(interaction=data['interaction'],
                  direction=data['direction'],
                  correspondent_id=data['correspondent_id'],
                  datetime=_parse_date(data['datetime']),
                  call_duration=_map_duration(data['call_duration']),
                  position=_map_position(data))


def filter_record(records):
    """
    Filter records and remove items with missing or inconsistents fields

    Parameters
    ----------

    records : list
        A list of Record objects

    Returns
    -------

    records, ignored : (Record list, dict)
        A tuple of filtered records, and a dictionnary counting the missings fields

    """

    scheme = {
        'interaction': lambda r: r.interaction in ['call', 'text'],
        'direction': lambda r: r.direction in ['in', 'out'],
        'correspondent_id': lambda r: r.correspondent_id is not None,
        'datetime': lambda r: isinstance(r.datetime, datetime),
        'call_duration': lambda r: isinstance(r.call_duration, (int, float)) if r.interaction == 'call' else True,
    }

    ignored = OrderedDict([
        ('all', 0),
        ('interaction', 0),
        ('direction', 0),
        ('correspondent_id', 0),
        ('datetime', 0),
        ('call_duration', 0),
    ])

    def _filter(records):
        global removed

        for r in records:
            valid = True
            for key, test in scheme.iteritems():
                if not test(r):
                    ignored[key] += 1
                    valid = False  # Not breaking, we count all fields with errors

            if valid is True:
                yield r
            else:
                ignored['all'] += 1

    return list(_filter(records)), ignored


def load(name, records, antennas, attributes=None, antennas_path=None,
         attributes_path=None, describe=True, warnings=False):
    """
    Creates a new user. This function is used by read_csv, read_orange,
    and read_telenor. If you want to implement your own reader function, we advise you to use the load() function

    `load` will output warnings on the standard output if some records or
    antennas are missing a position.

    Parameters
    ----------

    name : str
        The name of the user. It is stored in User.name and is useful when
        exporting metrics about multiple users.

    records: list
        A list or a generator of Record objects.

    antennas : dict
        A dictionary of the position for each antenna.

    attributes : dict
        A (key,value) dictionary of attributes for the current user

    describe : boolean
        If describe is True, it will print a description of the loaded user
        to the standard output.

    warnings : boolean, default True
        If warnings is equal to False, the function will not output the
        warnings on the standard output.


    For instance:

    .. code-block:: python

       >>> records = [Record(...),...]
       >>> antennas = {'A51': (37.245265, 115.803418),...}
       >>> attributes = {'age': 60}
       >>> load("Frodo", records, antennas, attributes)

    will returns a new User object.


    """

    user = User()
    user.name = name
    user.antennas_path = antennas_path
    user.attributes_path = attributes_path

    user.records, ignored = filter_record(records)

    if ignored['all'] != 0:
        if warnings:
            print warning_str("Warning: %d record(s) were removed due to missing or incomplete fields." % ignored['all'])
        for k in ignored.keys():
            if k != 'all' and warnings:
                print warning_str(" " * 9 + "%s: %i record(s) with incomplete values" % (k, ignored[k]))

    user.ignored_records = dict(ignored)

    if antennas is not None:
        user.antennas = antennas
        user.has_antennas = True
    if attributes is not None:
        user.attributes = attributes

    percent_missing = percent_records_missing_location(user)
    if percent_missing > 0:
        if warnings:
            print warning_str("Warning: {0:.2%} of the records are missing a location.".format(percent_missing))
        if antennas is None and warnings:
            print warning_str("         No antennas file was given and records are using antennas for position")

    if antennas_missing_locations(user) > 0:
        if warnings:
            print warning_str("Warning: %d antenna(s) are missing a location." % antennas_missing_locations(user))

    if describe is True:
        user.describe()

    return user


def _record_match(ours, theirs, time_diff=30):
    """
    Return True if two records match.
    """
    return ours.interaction == theirs.interaction and \
        ours.direction != theirs.direction and \
        ours.call_duration == theirs.call_duration and \
        abs((ours.datetime - theirs.datetime).total_seconds()) < time_diff


def _read_network(user, records_path, attributes_path, read_function, antennas_path=None, extension=".csv"):
    connections = {}
    correspondents = Counter([record.correspondent_id for record in user.records])

    num_records_error = 0

    # Try to load all the possible correspondent files
    for c_id, count in sorted(correspondents.items()):
        correspondent_file = os.path.join(records_path, c_id + extension)
        if os.path.exists(correspondent_file):
            correspondent_user = read_function(c_id, records_path, antennas_path, attributes_path, describe=False, network=False)

            # Look for matching record in the correspondent
            for our_record in filter(lambda x: x.correspondent_id == c_id, user.records):
                for their_record in correspondent_user.records:
                    # See if they match
                    if _record_match(our_record, their_record):
                        break
                else:
                    num_records_error += 1

        else:
            correspondent_user = None

        connections[c_id] = correspondent_user

    if len(user.records) > 0 and num_records_error > 0:
        percent_inconsistant = num_records_error / len(user.records)
        print warning_str('Warning: {} records of the current user are not reciprocated ({:.2%}).'.format(num_records_error, percent_inconsistant))

    # Return the network dictionary sorted by key
    return OrderedDict(sorted(connections.items(), key=lambda t: t[0]))


def read_csv(user_id, records_path, antennas_path=None, attributes_path=None, network=True, describe=True, warnings=True):
    """
    Load user records from a CSV file.

    Parameters
    ----------

    user_id : str
        ID of the user (filename)

    records_path : str
        Path of the directory all the user files.

    antennas_path : str, optional
        Path of the CSV file containing (place_id, latitude, longitude) values.
        This allows antennas to be mapped to their locations.

    attributes_path : str, optional
        Path of the directory containing attributes files (``key, value`` CSV file).
        Attributes can for instance be variables such as like, age, or gender.
        Attributes can be helpful to compute specific metrics.

    network : bool, optional
        If network is True, bandicoot loads the network of the user's correspondants from the same path. Defaults to False.

    describe : boolean
        If describe is True, it will print a description of the loaded user to the standard output.


    Examples
    --------

    >>> user = bandicoot.read_csv('sample_records', '.')
    >>> print len(user.records)
    10

    >>> user = bandicoot.read_csv('sample_records', 'samples', sample_places.csv')
    >>> print len(user.antennas)
    5

    >>> user = bandicoot.read_csv('sample_records', '.', None, 'sample_attributes.csv')
    >>> print user.attributes['age']
    25

    Notes
    -----
    - The csv files can be single, or double quoted if needed.
    - Empty cells are filled with ``None``. For example, if the column
      ``call_duration`` is empty for one record, its value will be ``None``.
      Other values such as ``"N/A"``, ``"None"``, ``"null"`` will be
      considered as a text.
    """

    antennas = None
    if antennas_path is not None:
        with open(antennas_path, 'rb') as csv_file:
            reader = csv.DictReader(csv_file)
            antennas = dict((d['place_id'], (float(d['latitude']),
                                             float(d['longitude'])))
                            for d in reader)

    user_records = os.path.join(records_path, user_id + '.csv')
    with open(user_records, 'rb') as csv_file:
        reader = csv.DictReader(csv_file)
        records = map(_parse_record, reader)

    attributes = None
    if attributes_path is not None:
        user_attributes = os.path.join(attributes_path, user_id + '.csv')
        try:
            with open(user_attributes, 'rb') as csv_file:
                reader = csv.DictReader(csv_file)
                attributes = dict((d['key'], d['value']) for d in reader)
        except IOError:
            attributes = None

    user = load(user_id, records, antennas, attributes, antennas_path,
                attributes_path, describe, warnings)

    # Loads the network
    if network is True:
        user.network = _read_network(user, records_path, attributes_path, read_csv, antennas_path)
        user.recompute_missing_neighbors()

    return user


def read_orange(records_path=None, network=False, describe=True, warnings=True):
    """
    Load user records from a CSV file in *orange* format:

    ``call_record_type;basic_service;user_msisdn;call_partner_identity;datetime;call_duration;longitude;latitude``

    ``basic_service`` takes one of the following values:

    - 11: telephony;
    - 12: emergency calls;
    - 21: short message (in)
    - 22: short message (out)

    Parameters
    ----------

    records_path : str or iterator, optional
        If ``records_path`` is a string, the function will load the CSV file at
        the path ``records_path``. If the parameter is an iterator, records will
        be loaded directly, as an ordered list in the *orange* format.If no
        parameter is included, ``read_orange`` will load records from the
        standard input ``sys.stdin``

    describe : boolean
        If describe is True, it will print a description of the loaded user to the standard output.

    network : bool, optional
        If network is True, bandicoot loads the network of the user's correspondants from the same path. Defaults to False.
    """

    def _parse(reader):
        records = []
        antennas = dict()

        for row in reader:
            direction = 'out' if row[0] == '1' else 'in'
            interaction = 'call' if row[1] in ['11', '12'] else 'text'
            contact = row[3]
            date = datetime.strptime(row[4], "%Y-%m-%d %H:%M:%S")
            call_duration = float(row[5])
            lon, lat = float(row[6]), float(row[7])
            latlon = (lat, lon)

            antenna = None
            for key, value in antennas.items():
                if latlon == value:
                    antenna = key
                    break
            if antenna is None:
                antenna = len(antennas) + 1
                antennas[antenna] = latlon

            position = Position(antenna=antenna, location=latlon)

            record = Record(direction=direction,
                            interaction=interaction,
                            correspondent_id=contact,
                            call_duration=call_duration,
                            datetime=date,
                            position=position)
            records.append(record)

        return records, antennas

    name = records_path

    if isinstance(records_path, str):
        with open(records_path, 'rb') as f:
            reader = csv.reader(f, delimiter=";")
            records, antennas = _parse(reader)

    elif records_path is None:
        reader = csv.reader(sys.stdin, delimiter=";")
        records, antennas = _parse(reader)

    else:
        # Iteration over the first parameter
        records, antennas = _parse(records_path)

    user = load(name, records, antennas, warnings=None, describe=describe)

    if network is True:
        directory = os.path.dirname(records_path)
        user.network = _read_network(user, directory, read_orange)
        user.recompute_missing_neighbors()

    return user


def read_telenor(incoming_cdr, outgoing_cdr, cell_towers, describe=True, warnings=True):
    """
    Load user records from a CSV file in *telenor* format:

    Arguments
    ---------

    incoming_cdr: str
        Path to the CSV file containing incoming records, using the following
        scheme: ::

             B_PARTY,A_PARTY,DURATION,B_CELL,CALL_DATE,CALL_TIME,CALL_TYPE

    outgoing_cdr: str
        Path to the CSV file containing outgoing records, using the following
        scheme: ::

             A_NUMBER,B_NUMBER,DURATION,B_CELL,CALL_DATE,CALL_TIME,CALL_TYPE

    cell_towers: str
        Path to the CSV file containing the positions of all

    describe : boolean
        If describe is True, it will print a description of the loaded user to the standard output.

    """

    import itertools
    import csv

    def parse_direction(code):
        if code == 'MOC':
            return 'out'
        elif code == 'MTC':
            return 'in'
        else:
            raise NotImplementedError

    cells = None
    with open(cell_towers, 'rb') as f:
        cell_towers_list = csv.DictReader(f)
        cells = {}
        for line in cell_towers_list:
            if line['LONGITUDE'] != '' and line['LATITUDE'] != '':
                latlon = (float(line['LONGITUDE']), float(line['LATITUDE']))
                cell_id = line['CELLID_HEX']
                cells[cell_id] = latlon

    def parse_record(raw):
        direction = parse_direction(raw['CALL_TYPE'].strip())

        if direction == 'in':
            contact = raw.get('A_PARTY', raw.get('A_NUMBER'))
            cell_id = raw['B_CELL']
        else:
            contact = raw.get('B_PARTY', raw.get('B_NUMBER'))
            cell_id = raw['A_CELL']

        position = Position(antenna=cell_id, location=cells.get(cell_id))

        _date_str = raw.get('CDATE', raw.get('CALL_DATE'))
        _time_str = raw.get('CTIME', raw.get('CALL_TIME'))
        _datetime = datetime.strptime(_date_str + _time_str,
                                      "%Y%m%d%H:%M:%S")

        r = Record(interaction='call',
                   direction=direction,
                   correspondent_id=contact,
                   call_duration=float(raw['DURATION'].strip()),
                   datetime=_datetime,
                   position=position)

        return r

    with open(incoming_cdr, 'rb') as f_in:
        incoming_ = map(parse_record, csv.DictReader(f_in))

        with open(outgoing_cdr, 'rb') as f:
            outgoing_ = map(parse_record, csv.DictReader(f))

            records = itertools.chain(incoming_, outgoing_)

    name = incoming_cdr

    user = load(name, records, cells, warnings=None, describe=describe)
    return user



import bandicoot as bc
from os.path import dirname, abspath, join

if __name__ == '__main__':
    empty_user = bc.User()
    empty_user.attributes['empty'] = True
    empty_path = join(dirname(abspath(__file__)), 'samples/empty_user.json')
    bc.io.to_json(bc.utils.all(empty_user, summary='extended', flatten=True), empty_path)

    sample_user = bc.tests.generate_user.sample_user()
    sample_path = join(dirname(abspath(__file__)), 'samples/sample_user_all_metrics.json')
    bc.io.to_json(bc.utils.all(sample_user, summary='extended', groupby=None, flatten=True), sample_path)



"""
Test manual user files, made by real humans, with all metrics handcrafted.
"""

import bandicoot as bc
import unittest
from testing_tools import parse_dict, metric_suite
import os


class TestManual(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dir_changed = False

    def setUp(self):
        if not TestManual._dir_changed:
            abspath = os.path.abspath(__file__)
            name = abspath.index(os.path.basename(__file__))
            abspath = abspath[:name]
            os.chdir(abspath)
            TestManual._dir_changed = True

        self.user_A = bc.io.read_csv("A", "samples/manual", "samples/towers.csv", describe=False)
        self.user_B = bc.io.read_csv("B", "samples/manual", describe=False)

    def test_A_metrics(self):
        self.assertTrue(*metric_suite(self.user_A, parse_dict("samples/manual/A.json"), groupby=None, decimal=4))

    def test_B_metrics(self):
        self.assertTrue(*metric_suite(self.user_B, parse_dict("samples/manual/B.json"), groupby=None, decimal=3))


from __future__ import division

import datetime
from bandicoot.helper.fixes import Counter
from bandicoot.helper.tools import Colors
from bandicoot.helper.group import _binning
import bandicoot as bc


class Record(object):
    """
    Data structure storing a call detail record.

    Attributes
    ----------
    interaction : str, 'text' or 'call'
        The type of interaction (text message or call).
    direction : str, 'in' or 'out'
        The direction of the record (incoming or outgoing).
    correspondent_id : str
        A unique identifier for the corresponding contact
    datetime : datetime
        The exact date and time of the interaction
    call_duration : int or None
        Durations of the call in seconds. None if the record is a text message.
    position : Position
        The geographic position of the user at the time of the interaction.
    """

    __slots__ = ['interaction', 'direction', 'correspondent_id', 'datetime', 'call_duration', 'position']

    def __init__(self, interaction, direction, correspondent_id, datetime, call_duration, position):
        self.interaction = interaction
        self.direction = direction
        self.correspondent_id = correspondent_id
        self.datetime = datetime
        self.call_duration = call_duration
        self.position = position

    def __repr__(self):
        return "Record(" + ", ".join(map(lambda x: "%s=%r" % (x, getattr(self, x)), self.__slots__)) + ")"

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.__slots__ == other.__slots__:
            return all(getattr(self, attr) == getattr(other, attr) for attr in self.__slots__)
        return False


class Position(object):
    """
    Data structure storing a generic location. Can be instantiated with either an
    antenna or a gps location. Printing out the position will show which was used
    to instantiate it.
    """

    def __init__(self, antenna=None, location=None):
        self.antenna = antenna
        self.location = location

    def _get_location(self, user):
        if self.location:
            return self.location
        elif self.antenna:
            return user.antennas.get(self.antenna)
        else:
            return None

    def type(self):
        if self.antenna:
            return 'antenna'
        else:
            return 'gps'

    def __repr__(self):
        if self.antenna and self.location:
            return "Position(antenna=%s, location=%s)" % (self.antenna, self.location)
        if self.antenna:
            return "Position(antenna=%s)" % self.antenna
        if self.location:
            return "Position(location=(%s, %s))" % self.location

        return "Position()"

    def __eq__(self, other):
        if not isinstance(other, Position):
            return False
        if self.antenna and other.antenna:
            return self.antenna == other.antenna
        if self.location and other.location:
            return self.location == other.location

        if not self.location and not self.antenna and not other.location and not other.antenna:
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__repr__())


class User(object):
    """
    Data structure storing all the call, text or mobility records of the user.
    """

    def __init__(self):
        self._records = []
        self.name = None
        self.antennas_path = None
        self.attributes_path = None

        self.start_time = None
        self.end_time = None
        self.night_start = datetime.time(19)
        self.night_end = datetime.time(7)
        self.weekend = [6, 7]  # Saturday, Sunday by default

        self.home = None
        self.has_home = False
        self.has_text = False
        self.has_call = False
        self.has_antennas = False
        self.attributes = {}
        self.antennas = {}
        self.ignored_records = None

        self.percent_outofnetwork_calls = 0
        self.percent_outofnetwork_texts = 0
        self.percent_outofnetwork_contacts = 0
        self.percent_outofnetwork_call_durations = 0

        self.network = {}

    @property
    def records(self):
        """
        Can be used to get or set the list of user's records.

        If the records are modified, the start and end time as well as the
        infered house properties of the object User are recomputed.
        """
        return self._records

    @records.setter
    def records(self, input):
        self._records = sorted(input, key=lambda r: r.datetime)
        if len(self._records) > 0:
            self.start_time = self._records[0].datetime
            self.end_time = self._records[-1].datetime

        # Reset all the states
        self.has_call = False
        self.has_text = False
        self.has_antennas = False

        for r in self._records:
            if r.interaction == 'text':
                self.has_text = True
            elif r.interaction == 'call':
                self.has_call = True

            if r.position.type() == 'antenna':
                self.has_antennas = True

        self.recompute_home()

    def recompute_missing_neighbors(self):
        """
        Recomputes statistics for missing users of the current user's
        network:

        - ``User.percent_outofnetwork_calls``
        - ``User.percent_outofnetwork_texts``
        - ``User.percent_outofnetwork_contacts``
        - ``User.percent_outofnetwork_call_durations``

        This function is automatically called from :meth:`~bandicoot.io.read_csv`
        when loading a network user.
        """

        oon_records = [r for r in self.records if self.network.get(r.correspondent_id, None) is None]
        num_oon_calls = len([r for r in oon_records if r.interaction == 'call'])
        num_oon_texts = len([r for r in oon_records if r.interaction == 'text'])
        num_oon_neighbors = len(set(x.correspondent_id for x in oon_records))
        oon_call_durations = sum([r.call_duration for r in oon_records if r.interaction == 'call'])

        num_calls = len([r for r in self.records if r.interaction == 'call'])
        num_texts = len([r for r in self.records if r.interaction == 'text'])
        total_neighbors = len(set(x.correspondent_id for x in self.records))
        total_call_durations = sum([r.call_duration for r in self.records if r.interaction == 'call'])

        def _safe_div(a, b, default):
            return a / b if b != 0 else default

        # We set the percentage at 0 if no event occurs
        self.percent_outofnetwork_calls = _safe_div(num_oon_calls, num_calls, 0)
        self.percent_outofnetwork_texts = _safe_div(num_oon_texts, num_texts, 0)
        self.percent_outofnetwork_contacts = _safe_div(num_oon_neighbors, total_neighbors, 0)
        self.percent_outofnetwork_call_durations = _safe_div(oon_call_durations, total_call_durations, 0)

    def describe(self):
        """
        Generate a short description of the object.

        Examples
        --------
        >>> import bandicoot as bc
        >>> user = bc.User()
        >>> user.records = bc.tests.generate_user.random_burst(5)
        >>> user.describe()
        [x] 5 records from 2014-01-01 10:41:00 to 2014-01-01 11:21:00
            5 contacts
        [x] 1 attribute

        Note
        ----
        The summary is directly sent to the standard output.
        """

        def format_int(name, n):
            if n == 0 or n == 1:
                return "%i %s" % (n, name[:-1])
            else:
                return "%i %s" % (n, name)

        empty_box = Colors.OKGREEN + '[ ]' + Colors.ENDC + ' '
        filled_box = Colors.OKGREEN + '[x]' + Colors.ENDC + ' '

        if self.start_time is None:
            print empty_box + "No records stored"
        else:
            print (filled_box + format_int("records", len(self.records)) +
                   " from %s to %s" % (self.start_time, self.end_time))

        nb_contacts = bc.individual.number_of_contacts(self, interaction='callandtext', groupby=None)
        nb_contacts = nb_contacts['allweek']['allday']['callandtext']
        if not nb_contacts:
            print empty_box + "No contacts"
        else:
            print filled_box + format_int("contacts", nb_contacts)

        if len(self.attributes) == 0:
            print empty_box + "No attributes stored"
        else:
            print filled_box + format_int("attributes", len(self.attributes))

        if len(self.antennas) == 0:
            print empty_box + "No antennas stored"
        else:
            print filled_box + format_int("antennas", len(self.antennas))

        if not self.has_home:
            print empty_box + "No home"
        else:
            print filled_box + "Has home"

        if not self.has_text:
            print empty_box + "No texts"
        else:
            print filled_box + "Has texts"

        if not self.has_call:
            print empty_box + "No calls"
        else:
            print filled_box + "Has calls"

    def recompute_home(self):
        """
        Return the antenna where the user spends most of his time at night.
        None is returned if there are no candidates for a home antenna
        """

        if self.night_start < self.night_end:
            night_filter = lambda r: self.night_end > r.datetime.time() > self.night_start
        else:
            night_filter = lambda r: not(self.night_end < r.datetime.time() < self.night_start)

        # Bin positions by chunks of 30 minutes
        candidates = list(_binning(filter(night_filter, self._records)))

        if len(candidates) == 0:
            self.home = None
            self.has_home = False
            return None
        else:
            self.home = Counter(candidates).most_common()[0][0]
            self.has_home = True

        return self.home

    @property
    def has_attributes(self):
        return len(self.attributes) != 0

    def set_home(self, new_home):
        """
        Sets the user's home. The argument can be a Position object or a
        tuple containing location data.
        """
        self.has_home = True
        if type(new_home) is Position:
            self.home = new_home

        elif type(new_home) is tuple:
            self.home = Position(location=new_home)

        else:
            self.home = Position(antenna=new_home)



from collections import OrderedDict as NativeOrderedDict
from functools import update_wrapper
import itertools
import inspect
import math
import json

try:
    from thread import get_ident as _get_ident
except ImportError:
    from dummy_thread import get_ident as _get_ident


try:
    next
except NameError:
    from bandicoot.helper.fixes import next


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        from bandicoot.core import User
        if isinstance(obj, User):
            return repr(obj)
        else:
            return json.JSONEncoder.default(self, obj)


class OrderedDict(NativeOrderedDict):
    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        s = json.dumps(self, cls=CustomEncoder, indent=4)
        return s.replace('"<', '<').replace('>"', '>')
    pass


def advanced_wrap(f, wrapper):
    """
    Wrap a decorated function while keeping the same keyword arguments
    """
    f_sig = list(inspect.getargspec(f))
    wrap_sig = list(inspect.getargspec(wrapper))

    # Update the keyword arguments of the wrapper
    if f_sig[3] is None or f_sig[3] == []:
        f_sig[3], f_kwargs = [], []
    else:
        f_kwargs = f_sig[0][-len(f_sig[3]):]

    for key, default in zip(f_kwargs, f_sig[3]):
        wrap_sig[0].append(key)
        wrap_sig[3] = wrap_sig[3] + (default, )

    wrap_sig[2] = None  # Remove kwargs
    src = "lambda %s: " % (inspect.formatargspec(*wrap_sig)[1:-1])
    new_args = inspect.formatargspec(
        wrap_sig[0], wrap_sig[1], wrap_sig[2], f_kwargs,
        formatvalue=lambda x: '=' + x)
    src += 'wrapper%s\n' % new_args

    decorated = eval(src, locals())
    return update_wrapper(decorated, f)


class Colors:
    """
    The `Colors` class stores six codes to color a string. It can be used to print
    messages inside a terminal prompt.

    Examples
    --------
    >>> print Colors.FAIL + "Error: it's a failure!" + Colors.ENDC

    Attributes
    ----------
    HEADER
        Header color.
    OKBLUE
        Blue color for a success message
    OKGREEN
        Green color for a success message
    WARNING
        Warning color (yellow)
    FAIL
        Failing color (red)
    ENDC
        Code to disable coloring. Always add it after coloring a string.

    """

    def __init__(self):
        pass

    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[32m'
    WARNING = '\033[33m'
    FAIL = '\033[31m'
    ENDC = '\033[0m'


def warning_str(str):
    """
    Return a colored string using the warning color (`Colors.WARNING`).
    """
    return Colors.WARNING + str + Colors.ENDC


def percent_records_missing_location(user, method=None):
    """
    Return the percentage of records missing a location parameter.
    """

    if len(user.records) == 0:
        return 0.

    missing_locations = sum([1 for record in user.records if record.position._get_location(user) is None])
    return float(missing_locations) / len(user.records)


def antennas_missing_locations(user, Method=None):
    unique_antennas = set([record.position.antenna for record in user.records if record.position.antenna is not None])
    return sum([1 for antenna in unique_antennas if user.antennas.get(antenna) is None])


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return itertools.izip(a, b)


def mean(data):
    """
    Return the arithmetic mean of ``data``.

    Examples
    --------

    >>> mean([1, 2, 3, 4, 4])
    2.8
    """

    if len(data) < 1:
        return None

    return float(sum(data)) / len(data)


def kurtosis(data):
    """
    Return the kurtosis for ``data``.
    """

    if len(data) == 0:
        return None

    num = moment(data, 4)
    denom = moment(data, 2) ** 2.

    return num / denom if denom != 0 else 0


def skewness(data):
    """
    Returns the skewness of ``data``.
    """

    if len(data) == 0:
        return None

    num = moment(data, 3)
    denom = moment(data, 2) ** 1.5

    return num / denom if denom != 0 else 0.


def std(data):
    if len(data) == 0:
        return None

    variance = moment(data, 2)
    return variance ** 0.5


def moment(data, n):
    if len(data) <= 1:
        return 0

    _mean = mean(data)
    return float(sum([(item - _mean) ** n for item in data])) / len(data)


def median(data):
    """
    Return the median of numeric data, unsing the "mean of middle two" method.
    If ``data`` is empty, ``0`` is returned.

    Examples
    --------

    >>> median([1, 3, 5])
    3.0

    When the number of data points is even, the median is interpolated:
    >>> median([1, 3, 5, 7])
    4.0
    """

    if len(data) == 0:
        return None

    data.sort()
    return float((data[len(data) / 2] + data[(len(data) - 1) / 2]) / 2.)


def minimum(data):
    if len(data) == 0:
        return None

    return float(min(data))


def maximum(data):
    if len(data) == 0:
        return None

    return float(max(data))


class SummaryStats(object):
    """
    Data structure storing a numeric distribution

    Attributes
    ----------
    mean : float
        Mean of the distribution.
    std : float
        The standard deviation of the distribution.
    min: float
        The minimum value of the distribution.
    max: float
        The max value of the distribution.
    median : float
        The median value of the distribution
    skewness : float
        The skewness of the distribution, measuring its asymmetry
    kurtosis : float
        The kurtosis of the distribution, measuring its "peakedness"
    distribution : list
        The complete distribution, as a list of floats

    Note
    ----

    You can generate a *SummaryStats* object using the
    :meth:`~bandicoot.helper.tools.summary_stats` function.

    """

    __slots__ = ['mean', 'std', 'min', 'max', 'median', 'skewness', 'kurtosis', 'distribution']

    def __init__(self, mean, std, min, max, median, skewness, kurtosis, distribution):
        self.mean, self.std, self.min, self.max, self.median, self.skewness, self.kurtosis, self.distribution = mean, std, min, max, median, skewness, kurtosis, distribution

    def __repr__(self):
        return "SummaryStats(" + ", ".join(map(lambda x: "%s=%r" % (x, getattr(self, x)), self.__slots__)) + ")"

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.__slots__ == other.__slots__:
            return all(getattr(self, attr) == getattr(other, attr) for attr in self.__slots__)
        return False


def summary_stats(data):
    """
    Returns a :class:`~bandicoot.helper.tools.SummaryStats` object containing informations on the given distribution.

    Example
    -------
    >>> summary_stats([0, 1])
    SummaryStats(mean=0.5, std=0.5, min=0.0, max=1.0, median=0.5, skewness=0.0, kurtosis=1.0, distribution=[0, 1])
    """

    if len(data) < 1:
        return SummaryStats(0., 0., 0., 0., 0., 0., 0., [])

    data.sort()
    _median = median(data)

    _mean = mean(data)
    _std = std(data)
    _minimum = minimum(data)
    _maximum = maximum(data)
    _kurtosis = kurtosis(data)
    _skewness = skewness(data)
    _distribution = data

    return SummaryStats(_mean, _std, _minimum, _maximum, _median, _skewness, _kurtosis, _distribution)


def entropy(data):
    """
    Compute the Shannon entropy, a measure of uncertainty.
    """

    if len(data) == 0:
        return None

    n = sum(data)

    _op = lambda f: f * math.log(f)
    return - sum(_op(float(i) / n) for i in data)


def great_circle_distance(pt1, pt2):
    r = 6371.

    delta_latitude = (pt1[0]-pt2[0])/180*math.pi
    delta_longitude = (pt1[1]-pt2[1])/180*math.pi
    latitude1 = pt1[0]/180*math.pi
    latitude2 = pt2[0]/180*math.pi

    return r*2.*math.asin(math.sqrt(math.pow(math.sin(delta_latitude/2), 2) + math.cos(latitude1)*math.cos(latitude2) * math.pow(math.sin(delta_longitude/2), 2)))


class AutoVivification(dict):
    """
    Implementation of perl's autovivification feature.

    Under CC-BY-SA 3.0 from nosklo on stackoverflow:
    http://stackoverflow.com/questions/19189274/defaultdict-of-defaultdict-nested
    """
    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value



"""
Test the import of CSV files.
"""

import bandicoot as bc
from bandicoot.core import Record, Position
import unittest
from StringIO import StringIO
import sys
import datetime
import csv
import os

class TestParsers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dir_changed = False

    def setUp(self):
        if not TestParsers._dir_changed:
            abspath = os.path.abspath(__file__)
            name = abspath.index(os.path.basename(__file__))
            abspath = abspath[:name]
            os.chdir(abspath)
            TestParsers._dir_changed = True

        self.stdin = StringIO("""1;22;770000000;770000005;2013-12-16 07:30:30;0;42.3987;-71.5750
            1;11;770000000;770000008;2013-12-16 08:30:30;126;42.1534;-70.7084
            2;22;770000000;770000007;2013-12-16 17:30:30;0;42.6590;-71.1216""")
        sys.stdin = self.stdin

    def tearDown(self):
        sys.stdin = sys.__stdin__

    def test_read_orange(self):
        user = bc.io.read_orange("samples/u_test.csv", describe=False)
        self.assertEqual(len(user.records), 500)

    def test_read_orange_stdin(self):
        user = bc.io.read_orange(describe=False)
        self.assertEqual(user.records[0],
                         Record(interaction='text',
                                direction='out',
                                correspondent_id='770000005',
                                datetime=datetime.datetime(2013, 12, 16, 7, 30, 30),
                                call_duration=0.0,
                                position=Position(1, (42.3987, -71.575))))

    def test_read_orange_from_reader(self):
        with open("samples/u_test.csv") as f:
            r = csv.reader(f, delimiter=';')
            user = bc.io.read_orange(r, describe=False)
        self.assertEqual(len(user.records), 500)

    def test_read_csv(self):
        user = bc.read_csv("u_test2", "samples", describe=False)
        self.assertEqual(len(user.records), 500)

    def test_read_csv_antenna_id_no_places(self):
        user = bc.read_csv("u_test_antennas", "samples", describe=False)
        self.assertEqual(user.records[1],
                         Record(interaction='call',
                                direction='in',
                                correspondent_id='770000001',
                                datetime=datetime.datetime(2013, 12, 16, 5, 39, 30),
                                call_duration=0,
                                position=Position('13084', None)))

        result = {'allweek': {'allday': None}}
        self.assertEqual(bc.spatial.radius_of_gyration(user, groupby=None), result)

    def test_read_csv_antenna_id(self):
        user = bc.read_csv("u_test_antennas", "samples", antennas_path="samples/towers.csv", describe=False)
        self.assertEqual(user.records[1],
                         Record(interaction='call',
                                direction='in',
                                correspondent_id='770000001',
                                datetime=datetime.datetime(2013, 12, 16, 5, 39, 30),
                                call_duration=0,
                                position=Position('13084', None)))

        radius = bc.spatial.radius_of_gyration(user, groupby=None)['allweek']['allday']
        self.assertGreater(radius, 0)

    def test_read_csv_no_position(self):
        user = bc.read_csv("u_test_no_position", "samples", describe=False)
        self.assertEqual(user.records[1],
                         Record(interaction='call',
                                direction='in',
                                correspondent_id='770000001',
                                datetime=datetime.datetime(2013, 12, 16, 5, 39, 30),
                                call_duration=0,
                                position=Position()))

    def test_read_csv_attributes(self):
        user = bc.read_csv("u_test2", "samples", attributes_path="samples/attributes", describe=False)
        self.assertEqual(user.attributes, {
            'gender': 'male',
            'age': '42',
            'is_subscriber': 'True',
            'individual_id': '7atr8f53fg41'
        })



"""
Tests for the sample user
"""

import bandicoot as bc
import unittest
from testing_tools import parse_dict


class TestSampleUser(unittest.TestCase):
    def setUp(self):
        self.sample_user = bc.tests.generate_user.sample_user()

    def test_sample_user(self):
        result = bc.utils.all(self.sample_user, groupby=None, summary='extended', flatten=True)
        self.assertDictEqual(result, parse_dict("samples/sample_user_all_metrics.json")['sample_user'])



#!/usr/bin/python

import math

from bandicoot.helper.group import spatial_grouping
from bandicoot.helper.tools import entropy, great_circle_distance
from bandicoot.helper.fixes import Counter


@spatial_grouping(user_kwd=True)
def percent_at_home(positions, user):
    """
    The percentage of interactions the user had while he was at home.

    Notes
    -----
    The position of the home is computed by :meth:`User.compute_home`.
    If no home can be found, the percentage at home will be ``None``.
    """

    if not user.has_home:
        return None
    if user.home.antenna is None and user.home.location is None:
        return None
    positions = list(positions)
    total_home = sum(1 for p in positions if p == user.home)
    return float(total_home) / len(positions) if len(positions) != 0 else 0


@spatial_grouping
def radius_of_movement(records):
    return NotImplemented


@spatial_grouping(user_kwd=True)
def radius_of_gyration(positions, user):
    """
    The radius of gyration.

    The radius of gyration is the *equivalent distance* of the mass from the
    center of gravity, for all visited places  [GON2008]_

    None is returned if there are no (lat, lon) positions for where the user has been.

    .. [GON2008] Gonzalez, M. C., Hidalgo, C. A., & Barabasi, A. L. (2008). Understanding individual human mobility patterns. Nature, 453(7196), 779-782.

    """

    d = Counter(p._get_location(user) for p in positions if p._get_location(user) is not None)
    sum_weights = sum(d.values())
    positions = d.keys()  # Unique positions

    if len(positions) == 0:
        return None

    barycenter = [0, 0]
    for pos, t in d.items():
        barycenter[0] += pos[0] * t
        barycenter[1] += pos[1] * t

    barycenter[0] /= sum_weights
    barycenter[1] /= sum_weights

    r = 0.
    for pos, t in d.items():
        r += float(t) / sum_weights * great_circle_distance(barycenter, pos) ** 2
        r += float(t) / sum_weights * great_circle_distance(barycenter, pos) ** 2
    return math.sqrt(r)


@spatial_grouping
def entropy_of_antennas(positions):
    """
    The entropy of visited antennas.
    """
    counter = Counter(p for p in positions)
    return entropy(counter.values())


@spatial_grouping(use_records=True)
def number_of_antennas(records):
    """
    The number of unique places visited.
    """
    return len(set(r.position for r in records))


@spatial_grouping
def frequent_antennas(positions, percentage=0.8):
    """
    The number of location that account for 80% of the locations where the user was.
    """
    location_count = Counter(map(str, positions))

    target = int(math.floor(sum(location_count.values()) * percentage))
    location_sort = sorted(location_count.keys(),
                           key=lambda x: location_count[x])

    while target > 0 and len(location_sort) > 0:
        location_id = location_sort.pop()
        target -= location_count[location_id]

    return len(location_count) - len(location_sort)



from __future__ import division

from collections import Counter
from itertools import groupby, combinations
from functools import partial


def _count_interaction(user, interaction=None, direction='out'):
    if interaction is None:
        filtered = [x.correspondent_id for x in user.records if x.direction == direction]
    else:
        filtered = [x.correspondent_id for x in user.records if x.interaction == interaction and x.direction == direction]
    return Counter(filtered)


_count_text = partial(_count_interaction, interaction='text')
_count_call = partial(_count_interaction, interaction='call')


def _count_call_duration(user, direction='out'):
    """
    Returns a dictionary of (id, total duration of calls).
    """
    calls_out = (x for x in user.records if x.interaction == 'call' and x.direction == direction)
    grouped_by_id = groupby(calls_out, lambda r: r.correspondent_id)
    return {key: sum(r.call_duration for r in records) for (key, records) in grouped_by_id}


def __generate_matrix(user, generating_fn, default=0, missing=None):
    # Just in case, we remove the user from user.network (self records can happen)
    neighbors = [user.name] + sorted([k for k in user.network.keys() if k != user.name])

    def make_direction(direction):
        rows = []
        for u in neighbors:
            correspondent = user.network.get(u, user)

            if correspondent is None:
                row = [missing for v in neighbors]
            else:
                cur_out = generating_fn(correspondent, direction=direction)
                row = [cur_out.get(v, default) for v in neighbors]
            rows.append(row)
        return rows

    m1 = make_direction('out')
    m2 = make_direction('in')

    m = [[m1[i][j] or m2[j][i] for i in range(len(neighbors))] for j in range(len(neighbors))]
    return m


def _interaction_matrix(user, interaction=None):
    return __generate_matrix(user, partial(_count_interaction, interaction=interaction))


_interaction_matrix_call = lambda user: __generate_matrix(user, _count_call)
_interaction_matrix_text = lambda user: __generate_matrix(user, _count_text)
_interaction_matrix_call_duration = lambda user: __generate_matrix(user, _count_call_duration)


def clustering_coefficient(user, interaction=None, weighted=False):
    """
    The clustering coefficient of the user's ego undirected network.
    """

    matrix = _interaction_matrix(user, interaction=interaction)

    connected_triplets, triplets = 0, 0
    for a, b, c in combinations(range(len(matrix)), 3):
        if matrix[a][b] and matrix[b][c] and matrix[a][c]:
            if weighted:
                triplets += (matrix[a][b] * matrix[b][c]) ** .5
            elif not weighted:
                triplets += 1

            if matrix[a][b] != 0 and matrix[b][c] != 0 and matrix[a][c] != 0:
                if weighted:
                    connected_triplets += (matrix[a][b] * matrix[b][c]) ** .5
                else:
                    connected_triplets += 1

    return float(connected_triplets) / triplets if triplets != 0 else 0
    


"""
Test for the bandicoot.helper.group module.
"""

import bandicoot as bc
from bandicoot.core import Record, Position
import unittest
import datetime
from bandicoot.tests.generate_user import random_burst
from bandicoot.helper.group import group_records
from bandicoot.helper.tools import std, mean, minimum, maximum
from datetime import timedelta
import numpy as np
import os


class TestGroup(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dir_changed = False

    def setUp(self):
        if not TestGroup._dir_changed:
            abspath = os.path.abspath(__file__)
            name = abspath.index(os.path.basename(__file__))
            abspath = abspath[:name]
            os.chdir(abspath)
            TestGroup._dir_changed = True

        self.user = bc.io.read_orange("samples/u_test.csv", describe=False)
        self.random_int_list = np.random.randint(1, 1000, size=9001)

        self.sum_stats_list = [bc.helper.tools.SummaryStats(np.random.rand(), np.random.rand(),
                               np.random.rand(), np.random.rand(), np.random.rand(), np.random.rand(), np.random.rand(), []) for _ in range(9001)]

    def test_statistics(self):
        self.assertDictEqual(bc.helper.group.statistics(self.random_int_list, summary='default'), {
            'mean': mean(self.random_int_list),
            'std': std(self.random_int_list),
        })

        def mean_std(key):
            return {
                'mean': mean([getattr(s, key) for s in self.sum_stats_list]),
                'std': std([getattr(s, key) for s in self.sum_stats_list]),
            }

        self.assertDictEqual(bc.helper.group.statistics(self.sum_stats_list, summary='extended'), {
            'kurtosis': mean_std('kurtosis'),
            'mean': mean_std('mean'),
            'median': mean_std('median'),
            'skewness': mean_std('skewness'),
            'std': mean_std('std'),
            'min': mean_std('min'),
            'max': mean_std('max')
        })

        self.assertEqual(bc.helper.group.statistics([]).values(), [None] * 2)

    def test_statistics_bad_aggregated(self):
        def run_bad_aggregated():
            try:
                bc.helper.group.statistics("bad_aggregated")
            except (TypeError, ValueError):
                return True
            return False
        self.assertTrue(run_bad_aggregated())

    def test_weekly_group(self):
        records = [
            Record("test_itr", "in", "1", datetime.datetime(2014, 8, 24), 1, Position()),
            Record("test_itr", "in", "1", datetime.datetime(2014, 9, 4), 1, Position()),
            Record("test_itr", "in", "1", datetime.datetime(2014, 9, 11), 1, Position())
        ]
        user = bc.User()
        user.records = records

        grouping = bc.helper.group.group_records(user, groupby='week')
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[0].datetime)
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[1].datetime)
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[2].datetime)

    def test_weekday_group(self):
        records = [
            Record("test_itr", "in", "1", datetime.datetime(2014, 8, 25), 1, Position()),
            Record("test_itr", "in", "1", datetime.datetime(2014, 9, 4), 1, Position()),
            Record("test_itr", "in", "1", datetime.datetime(2014, 9, 11), 1, Position())
        ]
        user = bc.User()
        user.records = records

        grouping = bc.helper.group.group_records(user, groupby='week', part_of_week='weekday')
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[0].datetime)
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[1].datetime)
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[2].datetime)

    def test_weekend_group(self):
        records = [
            Record("test_itr", "in", "1", datetime.datetime(2014, 8, 23), 1, Position()),
            Record("test_itr", "in", "1", datetime.datetime(2014, 8, 31), 1, Position()),
            Record("test_itr", "in", "1", datetime.datetime(2014, 10, 18), 1, Position())
        ]
        user = bc.User()
        user.records = records

        grouping = bc.helper.group.group_records(user, groupby='week', part_of_week='weekend')
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[0].datetime)
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[1].datetime)
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[2].datetime)

    def test_none_group(self):
        records = [
            Record("call", "in", "1", datetime.datetime(2014, 9, 5), 1, Position()),
            Record("call", "in", "1", datetime.datetime(2014, 9, 4), 1, Position()),
            Record("call", "in", "1", datetime.datetime(2014, 9, 11), 1, Position()),
            Record("call", "in", "1", datetime.datetime(2014, 9, 12), 1, Position())
        ]
        user = bc.User()
        user.records = records

        grouping = bc.helper.group.group_records(user, groupby=None)
        record = grouping.next()
        self.assertTrue(record.next().datetime, records[0].datetime)
        self.assertTrue(record.next().datetime, records[1].datetime)
        self.assertTrue(record.next().datetime, records[2].datetime)
        self.assertTrue(record.next().datetime, records[3].datetime)
        self.assertRaises(StopIteration, record.next)
        self.assertRaises(StopIteration, grouping.next)


class ConsistencyTests(unittest.TestCase):
    def setUp(self):
        self.user = bc.User()
        self.user.records = random_burst(100, delta=timedelta(days=2))

    def _group_set(self, method, interaction):
        chunks = group_records(self.user, groupby='method',
                               interaction=interaction)
        new_records = set(r for c in chunks for r in c)
        return new_records

    def test_weekly(self):
        old_records = set(self.user.records)
        new_records = self._group_set('week', None)
        self.assertSetEqual(new_records, old_records)

        new_records = self._group_set('week', 'call')
        self.assertSetEqual(new_records, {r for r in old_records
                                          if r.interaction == 'call'})


class MissingTests(unittest.TestCase):
    def setUp(self):
        self.user = bc.read_csv('user_ignored', 'samples')

    def test_amount(self):
        result = {
            'all': 4,
            'interaction': 2,
            'direction': 2,
            'correspondent_id': 0,
            'datetime': 0,
            'call_duration': 1,
        }
        self.assertDictEqual(self.user.ignored_records, result)

    def test_total_records(self):
        self.assertEqual(len(self.user.records), 1)



#!/usr/bin/env python

from setuptools import setup
import re
import ast


_version_re = re.compile(r'__version__\s+=\s+(.*)')
with open('bandicoot/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))


setup(
    name='bandicoot',
    author='Yves-Alexandre de Montjoye',
    author_email='yvesalexandre@demontjoye.com',
    version=version,
    url="https://github.com/yvesalexandre/bandicoot",
    license="MIT",
    packages=[
        'bandicoot',
        'bandicoot.helper',
        'bandicoot.tests'
    ],
    description="A toolbox to analyze mobile phone metadata.",
    classifiers=[
        'Environment :: Plugins',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics'
    ])



#!/usr/bin/python

import hashlib
import datetime
import random
import math
import csv
import copy

from bandicoot.core import Record, Position
import bandicoot as bc
from bandicoot.helper.tools import OrderedDict


def random_record(**kwargs):
    n_users = 48
    rate = 1e-4

    year  = 2012
    month = random.choice(range(1, 3))
    day   = random.choice(range(1, 28))
    
    # ensures that some correspondents have more interactions than others
    correspondent = random.randint(0, n_users/2)+random.randint(0, n_users/2)

    r = {'datetime': datetime.datetime(year, month, day) + datetime.timedelta(seconds=math.floor(-1/rate*math.log(random.random()))),
         'interaction':  random.choice(['text', 'text', 'text', 'call', 'call']),
         'correspondent_id': "correspondent_{}".format(correspondent),
         'direction': random.choice(['in', 'in', 'out']),
         'call_duration': random.randint(1, 1000),
         'position': Position(location=(random.uniform(-5, 5), random.uniform(-5, 5)))}
    if r['interaction'] == "text":
        r['call_duration'] = None

    r.update(kwargs)
    return Record(**r)


def sample_user(number_records=1482, seed=42, pct_in_network=0.8):
    old_state = random.getstate()
    random.seed(seed)

    towers = {701:(42.3555,-71.099541),
        702:(42.359039,-71.094595),
        703:(42.360481,-71.087321),
        704:(42.361013,-71.097868),
        705:(42.370849,-71.114613),
        706:(42.3667427,-71.1069847),
        707:(42.367589,-71.076537)}
    towers_position = [Position(antenna=k, location=v) for k, v in towers.items()]

    ego_records = [random_record(position=random.choice(towers_position)) for _ in xrange(number_records)]
    user = bc.io.load("sample_user", ego_records, towers, None, describe=False)
    
    # create network
    correspondents = set([record.correspondent_id for record in ego_records])
    between_user_records = {}
    correspondent_records = {}
    connections = {}
    
    n_in_network     = int(len(correspondents)*pct_in_network)
    if (n_in_network % 2 != 0):
        n_in_network = n_in_network - 1     
    in_network_correspondents = random.sample(correspondents, n_in_network)
    
    def reverse_records(records, current_owner):
        for r in records:
            r.direction = 'out' if r.direction == 'in' else 'in'  
            r.correspondent_id = current_owner
        return records
            
    # set records from ego
    for c_id in sorted(in_network_correspondents):
        reciprocal_records = filter(lambda r: r.correspondent_id == c_id, ego_records)
        reciprocal_records = reverse_records(copy.deepcopy(reciprocal_records), "ego")
        correspondent_records[c_id]  = reciprocal_records
    
    def generate_group_with_random_links(pct_users_in_group):
        n_in_group                  = int(len(correspondents)*pct_users_in_group)
        group                       = random.sample(non_grouped_correspondents, n_in_group)
        networkusers_group          = list()
        for user in group:
            if user in in_network_correspondents:
                networkusers_group.append(user)

        def CreatePair(source):        
            user_pair = [source, random.sample(group, 1)[0]]
            if user_pair[0] in non_grouped_correspondents:
                non_grouped_correspondents.remove(user_pair[0])
            if user_pair[1] in non_grouped_correspondents:
                non_grouped_correspondents.remove(user_pair[1])            
        
            extra_records = [random_record(position=random.choice(towers_position), interaction=random.choice(['text', 'call', 'call']), correspondent_id=user_pair[1]) for _ in xrange(random.randrange(25,150))]
            correspondent_records[user_pair[0]].extend(extra_records)
            if (user_pair[1] in in_network_correspondents):
                correspondent_records[user_pair[1]].extend(reverse_records(copy.deepcopy(extra_records), user_pair[0]))        
        
        # create pairs of users
        for i in range(len(networkusers_group)):
            CreatePair(networkusers_group[i])
            if random.choice(range(2)) == 0:
                CreatePair(networkusers_group[i])


    non_grouped_correspondents = copy.deepcopy(correspondents)
    for i in range(4):
        generate_group_with_random_links(pct_users_in_group=0.4-i*0.1)
        
    # create user object
    for c_id in sorted(correspondents):
        if (c_id in in_network_correspondents):
            correspondent_user =  bc.io.load(c_id, correspondent_records[c_id], towers, None, describe=False)
        else:
            correspondent_user = None
        connections[c_id] = correspondent_user
    
    # return the network dictionary sorted by key
    user.network = OrderedDict(sorted(connections.items(), key=lambda t: t[0]))
    user.recompute_missing_neighbors()

    random.setstate(old_state)
    return user


def write_new_user(filepath, n=1960):
    user = sample_user(n)

    schema = ['interaction', 'direction', 'correspondent_id', 'datetime', 'call_duration', 'antenna_id']
    with open(filepath, "wb") as new_user:
        w = csv.writer(new_user)
        w.writerow(schema)

        for record in user.records:
            w.writerow([record.position.antenna if x == 'antenna_id' else getattr(record, x) for x in schema])

    print "Finished writing new user to " + filepath
    write_answers(user)

def write_answers(user):
    bc.io.to_json([bc.utils.all(user, groupby='week', summary='default')], "samples/automatic/automatic_result.json")
    bc.io.to_json([bc.utils.all(user, groupby='week', summary='extended')], "samples/automatic/automatic_result_extended.json")
    bc.io.to_json([bc.utils.all(user, groupby=None, summary='default')], "samples/automatic/automatic_result_no_grouping.json")


def random_burst(count, delta=datetime.timedelta(minutes=10), **kwargs):

    first_date = datetime.datetime(2014, 01, 01, 10, 41)

    for i in range(count):
        _date = first_date + delta * i
        yield random_record(datetime=_date, **kwargs)


def random_records(n, antennas, number_of_users=150, ingoing=0.7, percent_text=0.3, rate=1e-4):
    current_date = datetime.datetime(2013, 01, 01, 00, 00, 00)
    results = []

    for _ in range(n):
        current_date += datetime.timedelta(seconds=math.floor(-1/rate*math.log(random.random())))
        interaction = "text" if random.random() < percent_text else "call"
        r = Record(
            interaction=interaction,
            direction='in' if random.random() < ingoing else 'out',
            correspondent_id=hashlib.md5(str(random.randint(1, number_of_users))).hexdigest(),
            datetime=str(current_date),
            call_duration=random.randint(1, 1000) if interaction == "call" else '',
            position=random.choice(antennas)
        )
        results.append(r._asdict())

    return results



"""
Tests for bandicoot.core (User, Position, and Record classes)
"""

import bandicoot as bc
import unittest
import datetime
from testing_tools import parse_dict
import os


class TestCore(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dir_changed = False

    def setUp(self):
        if not TestCore._dir_changed:
            abspath = os.path.abspath(__file__)
            name = abspath.index(os.path.basename(__file__))
            abspath = abspath[:name]
            os.chdir(abspath)
            TestCore._dir_changed = True

        self.user = bc.io.read_orange("samples/u_test.csv", describe=False)

        self.empty_user = bc.User()
        self.empty_user.attributes['empty'] = True

        self.user_nocturnal = bc.io.read_orange("samples/u_test.csv", describe=False)
        self.user_nocturnal.night_start = datetime.time(7, 0)
        self.user_nocturnal.night_end = datetime.time(19, 0)

        self.maxDiff = None

    def test_recompute_home_place(self):
        old_home = self.user.home
        new_home = self.user.recompute_home()

        self.assertEqual(old_home, new_home)
        self.assertEqual(old_home, self.user.home)

    def test_empty_user_all(self):
        result = bc.utils.all(self.empty_user, summary='extended', flatten=True)
        self.assertDictEqual(dict(result), parse_dict("samples/empty_user.json")['null'])

    def test_set_home(self):
        towers = parse_dict("samples/towers.json")
        towers = {key: tuple(value) for (key, value) in towers.items()}
        new_home = bc.core.Position(antenna=towers["1"])
        self.user.set_home(new_home)
        self.assertEqual(self.user.home, new_home)

        new_home = bc.core.Position(location=(42.3555368,-71.099507))
        self.user.set_home((42.3555368,-71.099507))
        self.assertEqual(self.user.home, new_home)


class TestTowers(unittest.TestCase):
    def setUp(self):
        self.user = bc.io.read_csv("A", "samples/manual/", "samples/towers.csv", describe=False)

    def testParse(self):
        towers = parse_dict("samples/towers.json")
        towers = {key: tuple(value) for (key, value) in towers.items()}

        self.assertDictEqual(self.user.antennas, towers)



from functools import partial
import itertools
from bandicoot.helper.tools import mean, std, SummaryStats, advanced_wrap, warning_str, AutoVivification


def group_records(user, interaction=None, groupby='week', part_of_week='allweek', part_of_day='allday'):
    """
    Group records by year and week number. This function is used by the
    ``@grouping`` decorator.

    Parameters
    ----------
    records : iterator
        An iterator over records

    groupby : {'week', None}, default 'week'
        * 'week': group all records by year and week;
        * None: records are not grouped. This is useful if you don't want to
          divide records in chunks.
    part_of_week : {'allweek', 'weekday', 'weekend'}, default 'allweek'
        * 'weekend': keep only the weekend records
        * 'weekday': keep only the weekdays records
        * 'allweek': use all the records
    part_of_day : {'allday', 'day', 'night'}, default 'allday'
        * 'day': keep only the records during the day
        * 'night': keep only the records during the night
        * 'allday': use all the records
    interaction : object
        The interaction to filter records:
        * "callandtext", for only callandtext;
        * a string, to filter for one type;
        * None, to use all records.
    """

    records = user.records

    def _group_date(records, _fun):
        for _, chunk in itertools.groupby(records, key=lambda r: _fun(r.datetime)):
            yield chunk

    if interaction == 'callandtext':
        records = filter(lambda r: r.interaction in ['call', 'text'], records)
    elif interaction is not None:
        records = filter(lambda r: r.interaction == interaction, records)

    if part_of_week == 'weekday':
        records = filter(lambda r: r.datetime.isoweekday() not in user.weekend, records)
    elif part_of_week == 'weekend':
        records = filter(lambda r: r.datetime.isoweekday() in user.weekend, records)
    elif part_of_week is not 'allweek':
        raise KeyError("{} is not a valid value for part_of_week. it should be 'weekday', 'weekend' or 'allweek'.".format(part_of_week))

    if user.night_start < user.night_end:
        night_filter = lambda r: user.night_end > r.datetime.time() > user.night_start
    else:
        night_filter = lambda r: not(user.night_end < r.datetime.time() < user.night_start)

    if part_of_day == 'day':
        records = filter(night_filter, records)
    elif part_of_day == 'night':
        records = filter(lambda r: not(night_filter(r)), records)
    elif part_of_day is not 'allday':
        raise KeyError("{} is not a valid value for part_of_day. It should be 'day', 'night' or 'allday'.".format(part_of_day))

    if groupby is None:
        return _group_date(records, lambda _: None)
    else:
        return _group_date(records, lambda d: d.isocalendar()[:2])


def statistics(data, summary='default', datatype=None):
    """
    Return statistics (mean, standard error, standard error and median, min and max) on data metrics.

    Examples
    --------
    Given a list of integers or floating point numbers,
    ``statistics`` computes the mean and standard error of the mean, and the min and max.

    >>> statistics([0, 1, 2, 3])
    {'mean': 1.5, 'std': 1.2910, 'min': 0, 'max': 3}

    Given a list of ``SummaryStats`` tuples, the function will
    returns the mean, standard error of the mean, min and max for each attribute
    of the tuples.
    """

    def _default_stats(agg):
        if agg is None or len(agg) == 0:
            return {'mean': None, 'std': None}
        else:
            # Some functions may return None values
            # It's better to filter them
            agg = filter(lambda x: x is not None, agg)
            return {'mean': mean(agg), 'std': std(agg)}

    def _stats_dict(v):
        return {key: _default_stats([getattr(s, key, None) for s in data]) for key in v}

    summary_keys = {
        'default': ['mean', 'std'],
        'extended': ['mean', 'std', 'median', 'skewness', 'kurtosis', 'min', 'max']
    }

    # Infer the data type of 'data'
    if datatype is None:
        if isinstance(data, (type(None), int, float)):
            datatype = 'scalar'
        elif isinstance(data, SummaryStats):
            datatype = 'summarystats'
        elif hasattr(data, "__len__"):  # list or numpy array
            data = filter(lambda x: x is not None, data)
            if len(data) == 0 or isinstance(data[0], (int, float)):
                datatype = 'distribution_scalar'
            elif isinstance(data[0], SummaryStats):
                datatype = 'distribution_summarystats'
            else:
                raise TypeError("{} is not a valid input. It should be a number, a SummaryStats object, or None".format(data[0]))
        else:
            raise TypeError("{} is not a valid input. It should be a number, a SummaryStats object, or a list".format(data))

    if datatype == 'scalar':
        return data

    if datatype == 'summarystats':
        if summary is None:
            return data.distribution
        elif summary in ['default', 'extended']:
            return {key: getattr(data, key, None) for key in summary_keys[summary]}
        else:
            raise ValueError("{} is not a valid summary type".format(summary))

    if datatype == 'distribution_scalar':
        if summary is 'default':
            return _default_stats(data)
        elif summary is None:
            return data
        else:
            raise ValueError("{} is not a valid summary type".format(summary))

    if datatype == 'distribution_summarystats':
        if summary is None:
            return [item.distribution for item in data]
        elif summary in ['extended', 'default']:
            return _stats_dict(summary_keys[summary])
        else:
            raise ValueError("{} is not a valid summary type".format(summary))

    else:
        raise ValueError("{} is not a valid data type.".format(datatype))


def grouping(f=None, user_kwd=False, interaction=['call', 'text'], summary='default'):
    """
    ``grouping`` is a decorator for indicator functions, used to simplify the source code.

    Parameters
    ----------
    f : function
        The function to decorate
    user_kwd : boolean
        If user_kwd is True, the user object will be passed to the decorated function
    interaction : 'call', 'text', 'location', or a list
        By default, all indicators use only 'call' and 'text' records, but the
        interaction keywords filters the records passed to the function.
    summary: 'default', 'extended', None
        An indicator returns data statistics, ether *mean* and *std* by
        default, more with 'extended', or the inner distribution with None.
        See :meth:`~bandicoot.helper.group.statistics` for more details.


    See :ref:`new-indicator-label` to learn how to write an indicator with this decorator.

    """

    if f is None:
        return partial(grouping, user_kwd=user_kwd, interaction=interaction, summary=summary)

    def wrapper(user, groupby='week', interaction=interaction, summary=summary, split_week=False, split_day=False, datatype=None, **kwargs):
        if interaction is None:
            interaction = ['call', 'text']
        elif isinstance(interaction, str):
            interaction = [interaction]
        else:
            interaction = interaction[:]  # copy the list for more safety

        part_of_day = ['allday']
        if split_day:
            part_of_day += ['day', 'night']

        part_of_week = ['allweek']
        if split_week:
            part_of_week += ['weekday', 'weekend']

        for i in interaction:
            if i not in ['callandtext', 'call', 'text', 'location']:
                raise ValueError("%s is not a valid interaction value. Only 'call', "
                                 "'text', and 'location' are accepted." % i)

        def map_filters(interaction, part_of_week, part_of_day):
            """
            Call the wrapped function for every combinations of interaction,
            part_of_week, and part_of_day.
            """
            for i in interaction:
                for filter_week in part_of_week:
                    for filter_day in part_of_day:
                        if user_kwd is True:
                            result = [f(g, user, **kwargs) for g in group_records(user, i, groupby, filter_week, filter_day)]
                        else:
                            result = [f(g, **kwargs) for g in group_records(user, i, groupby, filter_week, filter_day)]

                        yield filter_week, filter_day, i, result

        returned = AutoVivification()  # nested dict structure
        for (f_w, f_d, i, m) in map_filters(interaction, part_of_week, part_of_day):
            if groupby is None:
                m = m[0] if len(m) != 0 else None
            returned[f_w][f_d][i] = statistics(m, summary=summary, datatype=datatype)

        return returned

    return advanced_wrap(f, wrapper)


def _binning(records):
    """
    Bin records by chunks of 30 minutes, returning the most prevalent position.
    """

    records = list(records)

    from collections import Counter

    def get_key(d):
        from datetime import datetime, timedelta
        k = d + timedelta(minutes=-(d.minute % 30))
        return datetime(k.year, k.month, k.day, k.hour, k.minute, 0)

    chunks = itertools.groupby(records, key=lambda r: get_key(r.datetime))

    for _, items in chunks:
        positions = [i.position for i in items]
        yield Counter(positions).most_common(1)[0][0]


def spatial_grouping(f=None, user_kwd=False, summary='default', use_records=False):
    if f is None:
        return partial(spatial_grouping, user_kwd=user_kwd, summary=summary,
                       use_records=use_records)

    if use_records is True:
        map_records = lambda x: x
    else:
        map_records = _binning

    def wrapper(user, groupby='week', summary=summary, split_week=False, split_day=False, datatype=None, **kwargs):
        part_of_day = ['allday']
        if split_day:
            part_of_day += ['day', 'night']

        part_of_week = ['allweek']
        if split_week:
            part_of_week += ['weekday', 'weekend']

        def map_filters_spatial(part_of_week, part_of_day):
            """
            Call the wrapped function for every combinations of interaction,
            part_of_week, and part_of_day.
            """
            for filter_week in part_of_week:
                for filter_day in part_of_day:
                    if user_kwd is True:
                        result = [f(map_records(g), user, **kwargs) for g in group_records(user, None, groupby, filter_week, filter_day)]
                    else:
                        result = [f(map_records(g), **kwargs) for g in group_records(user, None, groupby, filter_week, filter_day)]

                    yield filter_week, filter_day, result

        returned = AutoVivification()  # nested dict structure
        for (f_w, f_d, m) in map_filters_spatial(part_of_week, part_of_day):
            if groupby is None:
                m = m[0] if len(m) != 0 else None
            returned[f_w][f_d] = statistics(m, summary=summary, datatype=datatype)

        return returned

    return advanced_wrap(f, wrapper)



#!/usr/bin/python

from bandicoot.helper.group import grouping
from bandicoot.helper.tools import summary_stats, entropy, pairwise
from bandicoot.helper.fixes import Counter, total_seconds

import math
import datetime
from collections import defaultdict


@grouping
def interevent_time(records):
    """
    The interevent time between two records of the user.
    """
    inter_events = pairwise(r.datetime for r in records)
    inter = [total_seconds(new - old) for old, new in inter_events]

    return summary_stats(inter)


@grouping
def number_of_contacts(records, more=0):
    """
    The number of contacts the user interacted with.
    """
    counter = Counter(r.correspondent_id for r in records)
    return sum(1 for d in counter.values() if d > more)


@grouping
def entropy_of_contacts(records):
    """
    The entropy of the user's contacts.
    """
    counter = Counter(r.correspondent_id for r in records)
    return entropy(counter.values())


@grouping
def interactions_per_contact(records):
    """
    The number of interactions a user had with each of its contacts.
    """
    counter = Counter(r.correspondent_id for r in records)
    return summary_stats(counter.values())


@grouping(user_kwd=True, interaction='call')
def percent_initiated_interactions(records, user):
    """
    The percentage of initiated interactions by the user.
    """
    records = list(records)

    if len(records) == 0:
        return 0

    initiated = sum(1 for r in records if r.direction == 'out')
    return float(initiated) / len(records)


@grouping(user_kwd=True)
def percent_nocturnal(records, user):
    """
    The percentage of interactions the user had at night.

    By default, nights are 7pm-7am. Nightimes can be set in
    ``User.night_start`` and ``User.night_end``.
    """
    records = list(records)

    if len(records) == 0:
        return 0

    if user.night_start < user.night_end:
        night_filter = lambda d: user.night_end > d.time() > user.night_start
    else:
        night_filter = lambda d: not(user.night_end < d.time() < user.night_start)

    return float(sum(1 for r in records if night_filter(r.datetime))) / len(records)


@grouping(interaction='call')
def call_duration(records):
    """
    The duration of the user's calls (in and out).
    """
    call_durations = [r.call_duration for r in records]
    return summary_stats(call_durations)


def _conversations(group, delta=datetime.timedelta(hours=1)):
    """
    Group texts into conversations. The function returns an iterator over records grouped by conversations.

    See :ref:`Using bandicoot <conversations-label>` for a definition of conversations.
    """
    last_time = None
    results = []
    for g in group:

        if last_time is None or g.datetime - last_time < delta:
            if g.interaction == 'text':
                results.append(g)

            # A call always ends a conversation
            else:
                if len(results) != 0:
                    yield results
                    results = []

        else:
            if len(results) != 0:
                yield results

            if g.interaction == 'call':
                results = []
            else:
                results = [g]

        last_time = g.datetime

    if len(results) != 0:
        yield results


@grouping(interaction='callandtext')
def response_rate_text(records):
    """
    The response rate of the user (between 0 and 1).

    The response rate is the percentage of conversations 1) which
    started with an incoming text and 2) in which the user sent at least one response.

    The following sequence of messages defines four conversations (``I`` for an
    incoming text, ``O`` for an outgoing text): ::

        I-O-I-O => Started with an incoming text and at least one outgoing text
        I-I-O-I => Started with an incoming text and at least one outgoing text
        I-I-I-I => Started with an incoming text but doesn't have outgoing texts
        O-O-I-O => Not starting with an incoming text

    Here, the ratio would be 2/3 as we have 3 conversations starting with an incoming text and 2 of them have at least one outgoing text.

    See :ref:`Using bandicoot <conversations-label>` for a definition of conversations.
    """
    records = list(records)

    interactions = defaultdict(list)
    for r in records:
        interactions[r.correspondent_id].append(r)

    def _response_rate(grouped):
        received, responded = 0, 0
        conversations = _conversations(grouped)

        for conv in conversations:
            if len(conv) != 0:
                first = conv[0]
                if first.direction == 'in' and first.interaction == 'text':
                    received += 1
                    if any((i.direction == 'out' for i in conv)):
                        responded += 1

        return responded, received

    # Group all records by their correspondent, and compute the response rate
    # for each
    all_couples = map(_response_rate, interactions.values())
    responded, received = map(sum, zip(*all_couples))

    return float(responded) / received if received != 0 else 0


@grouping(interaction='callandtext')
def response_delay_text(records):
    """
    The response delay of the user within a conversation (in seconds)

    The following sequence of messages defines conversations (``I`` for an
    incoming text, ``O`` for an outgoing text, ``-`` for a one minute
    delay): ::

        I-O--I----O, we have a 60 seconds response delay and a 240 seconds response delay
        O--O---I--O, we have a 1200 seconds response delay
        I--II---I-I, we don't have a response delay. The user hasn't answered

    For this user, the distribution of response delays will be ``[60, 240, 60]``

    Notes
    -----
    See :ref:`Using bandicoot <conversations-label>` for a definition of conversations.
    Conversation are defined by a serie of text messages all sent within an hour of one another. The response delay can thus not be higher than one hour.
    """

    records = list(records)

    interactions = defaultdict(list)
    for r in records:
        interactions[r.correspondent_id].append(r)

    def _response_delay(grouped):
        ts = (total_seconds(b.datetime - a.datetime)
              for conv in _conversations(grouped)
              for a, b in pairwise(conv)
              if b.direction == 'out' and a.direction == 'in')

        return ts

    delays = [r for i in interactions.values() for r in _response_delay(i)
              if r > 0]

    if delays == []:
        return None

    return summary_stats(delays)


@grouping(interaction='callandtext')
def percent_initiated_conversations(records):
    """
    The percentage of conversations that have been initiated by the user.

    See :ref:`Using bandicoot <conversations-label>` for a definition of conversations.
    """
    records = list(records)

    interactions = defaultdict(list)
    for r in records:
        interactions[r.correspondent_id].append(r)

    def _percent_initiated(grouped):
        mapped = [(1 if conv[0].direction == 'out' else 0, 1)
                  for conv in _conversations(grouped)]
        return mapped

    all_couples = [sublist for i in interactions.values()
                   for sublist in _percent_initiated(i)]

    if len(all_couples) == 0:
        init, total = 0, 0
    else:
        init, total = map(sum, zip(*all_couples))

    return float(init) / total if total != 0 else 0


@grouping(interaction='callandtext')
def active_days(records):
    """
    The number of days during which the user was active. A user is considered
    active if he sends a text, receives a text, initiates a call, receives a
    call, or has a mobility point.
    """

    days = set(r.datetime.date() for r in records)
    return len(days)


@grouping
def percent_pareto_interactions(records, percentage=0.8):
    """
    The number of user's contacts that account for 80% of its interactions.
    """

    records = list(records)
    if records == []:
        return 0.

    user_count = Counter(r.correspondent_id for r in records)

    target = int(math.floor(sum(user_count.values()) * percentage))
    user_sort = sorted(user_count.keys(), key=lambda x: user_count[x])

    while target > 0 and len(user_sort) > 0:
        user_id = user_sort.pop()
        target -= user_count[user_id]

    return (len(user_count) - len(user_sort)) / len(records)


@grouping(interaction='call')
def percent_pareto_durations(records, percentage=0.8):
    """
    The number of user's contacts that account for 80% of its total time spend on the phone.
    """

    records = list(records)
    if records == []:
        return 0.

    user_count = defaultdict(int)
    for r in records:
        if r.interaction == "call":
            user_count[r.correspondent_id] += r.call_duration

    target = int(math.floor(sum(user_count.values()) * percentage))
    user_sort = sorted(user_count.keys(), key=lambda x: user_count[x])

    while target > 0 and len(user_sort) > 0:
        user_id = user_sort.pop()
        target -= user_count[user_id]

    return (len(user_count) - len(user_sort)) / len(records)


@grouping
def balance_of_contacts(records, weighted=True):
    """
    The balance of interactions per contact. For every contact,
    the balance is the number of outgoing interactions divided by the total
    number of interactions (in+out).

    .. math::

       \\forall \\,\\text{contact}\\,c,\\;\\text{balance}\,(c) = \\frac{\\bigl|\\text{outgoing}\,(c)\\bigr|}{\\bigl|\\text{outgoing}\,(c)\\bigr|+\\bigl|\\text{incoming}\,(c)\\bigr|}

    Parameters
    ----------
    weighted : str, optional
        If ``True``, the balance for each contact is weighted by
        the number of interactions the user had with this contact.
    """

    counter_out = defaultdict(int)
    counter = defaultdict(int)

    for r in records:
        if r.direction == 'out':
            counter_out[r.correspondent_id] += 1
        counter[r.correspondent_id] += 1

    if not weighted:
        balance = [float(counter_out[c]) / float(counter[c]) for c in counter]
    else:
        balance = [float(counter_out[c]) / float(sum(counter.values())) for c in counter]

    return summary_stats(balance)


@grouping()
def number_of_interactions(records, direction=None):
    """
    The number of interactions.

    Parameters
    ----------
    direction : str, optional
        Filters the records by their direction: ``None`` for all records,
        ``'in'`` for incoming, and ``'out'`` for outgoing.
    """
    if direction is None:
        return len([r for r in records])
    else:
        return len([r for r in records if r.direction == direction])
